
#ifndef HH_HANDLER
#define HH_HANDLER

#include <cassert>
#include <string>
#include <vector>
#include <map>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>

#include "tesla.h"

using namespace std;

// address: a string-like structure *with* the same copy/free semantics for the
// buffer as string has
class address {
    string bits;

public:
    address() {}
    address(const void *addr, socklen_t len) : bits((const char *)(addr), len) {}

    // Allow default copy and assignment constructors (which just copy the string)

    const sockaddr *addr() const { return (const sockaddr *)(bits.data()); }
    socklen_t addrlen() const { return bits.length(); }

    const char *c_str() const {
	static char tmpbuf[1024];
	const sockaddr_in *addr = (const sockaddr_in *)bits.data();
	const char *buf = inet_ntoa(addr->sin_addr);
	if (buf == 0) buf = "???";
	sprintf(tmpbuf, "%s:%d", buf, ntohs(addr->sin_port));
	return tmpbuf;
    }
};


// data: an *un-owned* data/length pair.  Unlike string, has absolutely no
// semantics regarding the copying or freeing of the data; this is essential
// a pair<const char*, unsigned int>.
class data
{
private:
    const char *dbits;
    unsigned int dlength;
    const sockaddr *daddr;
    const socklen_t daddrlen;

public:
    data() : dbits(0), dlength(0), daddr(0), daddrlen(0) {}
    data(const char *str) : dbits(str), dlength(strlen(str)), daddr(0), daddrlen(0) {}
    data(const char *_dbits, unsigned int _dlength, const sockaddr *_daddr = 0, socklen_t _daddrlen = 0) :
	dbits(_dbits), dlength(_dlength), daddr(_daddr), daddrlen(_daddrlen) {}
    data(const string& s) : dbits(s.data()), dlength(s.length()), daddr(0), daddrlen(0) {}

    const char *bits() const { return dbits; }
    unsigned int length() const { return dlength; }
    const sockaddr *addr() const { return daddr; }
    socklen_t addrlen() const { return daddrlen; }

    operator string() const { return dlength == 0 ? string() : string(dbits, dlength); }
    const char & operator[](unsigned int index) const { return dbits[index]; }
};

class flow_handler;
class acceptret {
public:
    acceptret() : h(0) {}
    acceptret(flow_handler *_h, address _addr) : h(_h), addr(_addr) {}

    operator bool() const { return h; }

    flow_handler *const h;
    const address addr;
};

class flow_handler {
public:
    class init_context;
    typedef flow_handler *(*create_func)(init_context &);

    class info {
    private:
	string name;
	string clazz;
	int domain;
	int type;
	create_func cf;

	info(const info&);

    public:        
	info(const string& _name, const string& _clazz,
	     int _domain, int _type,
	     create_func _cf) :
	    name(_name), clazz(_clazz), domain(_domain), type(_type), cf(_cf)
	{
	}

	string get_name() const { return name; }
	string get_clazz() const { return clazz; }
	int get_domain() const { return domain; }
	int get_type() const { return type; }
	create_func get_cf() const { return cf; }
	flow_handler *create(init_context& ctxt) const {
	    ts_debug_1("In info::create; name=%s; class=%s", name.c_str(), clazz.c_str());
	    return (*cf)(ctxt);
	}
    };

    class init_context {
    private:
	const info& inf;
	flow_handler& upstream;
	int domain, type;
	int level;
	bool accepted;
	
    public:
	init_context(const info& _inf, flow_handler& _upstream, int _domain, int _type,
		     int _level, bool _accepted = false) :
	    inf(_inf), upstream(_upstream), domain(_domain), type(_type),
	    level(_level), accepted(_accepted) {
	}

	const info& get_info() const { return inf; }
	int get_domain() const { return domain; }
	int get_type() const { return type; }
	int get_level() const { return level; }
	flow_handler *plumb(flow_handler& me, int domain, int type) const { return flow_handler::plumb(me, domain, type, level); }
	bool is_accepted() const { return accepted; }

	flow_handler& get_upstream() const { return upstream; }
    };

private:
    int domain;
    int type;
    int level;
    const info& inf;

    // Prohibit copying
    template<class T> flow_handler& operator=(T&);

    typedef map<string, const info*> infomap;
    typedef vector<const info*> infovec;
    static infovec handlers;
    static infomap *reg_handlers;

protected:
    flow_handler* upstream;
    vector<flow_handler*> downstream;

    flow_handler(const flow_handler& h) : inf(h.get_info())
    {
	upstream = 0;
	domain = h.get_domain();
	type = h.get_type();
	level = h.get_level();
    }

    flow_handler(const init_context& ctxt, bool plumb = true);

    static void register_handler(const info&);

    virtual void dump_children(string& out, int indent) {
	for (vector<flow_handler*>::iterator i = downstream.begin(); i != downstream.end(); ++i)
	    (*i)->dump(out, indent);
    }

    flow_handler *plumb(int domain, int type, int fd = -1);

public:
    virtual ~flow_handler() {
	for (vector<flow_handler*>::iterator i = downstream.begin(); i != downstream.end(); ++i)
	    delete *i;
    }

    static void make_handlers();

    flow_handler& get_upstream() const { return *upstream; }
    void set_upstream(flow_handler *_upstream) { upstream = _upstream; }

    void dump(string& out, int indent = 0) {
	char ch[64];

	for (int i = 0; i < indent; ++i) {
	    out.append("   ");
	}
	out.append(" - ").append(get_info().get_name());
	out.append("; me=");
	sprintf(ch, "%p", this);
	out.append(ch).append("; upstream=");
	sprintf(ch, "%p", &get_upstream());
	out.append(ch).append("\n");

	dump_children(out, indent + 1);
    }

    const info& get_info() const { return inf; }
    int get_domain() const { return domain; }
    int get_type() const { return type; }
    int get_level() const { return level; }

    virtual int connect(address a);
    virtual int bind(address a);
    virtual acceptret accept();
    virtual int close();
    virtual void accept_ready(flow_handler *from);
    virtual void connected(flow_handler *from, bool success);
    virtual bool avail(flow_handler *from, data d);
    virtual int write(data d);
    virtual int shutdown(bool r, bool w);
    virtual int listen(int backlog);
    virtual address getsockname();
    virtual address getpeername();
    virtual void may_write(flow_handler *from, bool may);
    virtual void may_avail(bool may);

    virtual string getsockopt(int level, int optname, int maxlen);
    virtual int setsockopt(int level, int optname, string value);

    static flow_handler *plumb(flow_handler& upstream, int domain, int type, int level = 0);
};



#define DECLARE_HANDLER \
public: \
    static info _info; \
    static flow_handler *_create(init_context&); \
    static int _inited;

#define DECLARE_DUMMY_HANDLER \
public: \
    static info _info;

#define DEFINE_HANDLER(name, clazz, domain, type) \
flow_handler *clazz::_create(init_context& ctxt) { return new clazz(ctxt); } \
flow_handler::info clazz::_info(#name, #clazz, domain, type, clazz::_create); \
int clazz::_inited = (ts_debug_2("Registering flow_handler %s", #clazz), register_handler(_info), 1);

#define DEFINE_DUMMY_HANDLER(clazz) \
flow_handler::info clazz::_info(#clazz, #clazz, -1, -1, 0);





#endif
